//************************************************************************
// Example15HanoiGeneral.java
//
// We solve the following generalized version of the Hanoi Towers
// Problem: Initially, the disks are placed on the pegs in a given
// arbitrary way (which is legal - the disks are larger as we go down).
// We have to move them so as to achieve another given arbitrary
// configuration.
//
// The idea of the solution is to concentrate on the big problems
// first. Namely, we start by finding the largest disk not on the correct
// peg. We then find the largest disk which prevents us from moving the
// former (by being either on the same peg as the former or on the peg it
// needs to move to). This adds the mission of getting the latter disk
// out of the way. We continue by finding the largest obstruction to this
// move. Finally, we get a mission we can perform. We perform it and
// continue in the same way.
//
// The wishlist is held in a stack of type WishStack. The program runs by
// peeking at each stage at the wish on top of the stack, performing it
// if possible (and popping it from the stack) and adding the largest
// obstruction to this move otherwise.
//************************************************************************

class Example15HanoiGeneral
{
   public static void main (String[] args)
   {
      char[] start  = {'A', 'B', 'A', 'C'};
      char[] finish = {'B', 'C', 'C', 'A'};
      Configuration current  = new Configuration (start),
                    required = new Configuration (finish);
      for (int disk = current.getDisksNumber(); disk > 0; disk--)
      {
         WishStack wishes = new WishStack (disk);
         Wish wish = new Wish (disk, current.getPosition (disk),
                                     required.getPosition (disk));
         if (wish.isReal())
            wishes.push (wish);
         while (!wishes.isEmpty())
         {
            Wish currentWish = (Wish) wishes.peek();
            int  whichDisk = currentWish.getDisk();
            char from      = currentWish.getSource(),
                 to        = currentWish.getDestination();
            int nextDisk = whichDisk - 1;
            while (nextDisk > 0 && current.getPosition (nextDisk) != from
                                && current.getPosition (nextDisk) != to)
               nextDisk--;
            if (nextDisk == 0)
            {
               System.out.println ("move " + whichDisk + " from " + from
                                   + " to " + to);
               wishes.pop();
               current.setPosition (whichDisk, to);
            }
            else
            {
               char nextFrom = current.getPosition (nextDisk);
               char nextTo   = (char) ('A' + 'B' + 'C' - from - to);
               wishes.push (new Wish (nextDisk, nextFrom, nextTo));
            }
         }// while
      }// for
   }// method main
}// class Example15HanoiGeneral


class Wish
{
   int disk;
   char source, destination;

   Wish (int disk, char from, char to)
   {
      this.disk = disk;
      source = from;
      destination = to;
   }

   int getDisk()
   {return disk;}

   char getSource()
   {return source;}

   char getDestination()
   {return destination;}

   boolean isReal()
   {return destination != source;}
}// class Wish


class WishStack implements StackInterface
{
   private Wish[] wishes;
   private int size;

   WishStack (int maxSize)
   {
      wishes = new Wish[maxSize];
      size = 0;
   }// method WishStack

   public boolean push (Object wish)
   {
      boolean result = false;
      if (!isFull())
      {
         wishes[size] = (Wish) wish;
         size++;
         result = true;
      }
      return result;
   }// method push

   public Object pop()
   {
      size--;
      return wishes[size];
   }// method pop

   public Object peek()
   {return wishes [size - 1];}

   public boolean isEmpty()
   {return size == 0;}

   public boolean isFull()
   {return size == wishes.length;}
}// class WishStack


class Configuration
{
   private char[] where;

   Configuration (char[] where)
   {this.where = where;}

   char getPosition (int disk)
   {return where [disk - 1];}

   void setPosition (int disk, char peg)
   {where [disk - 1] = peg;}

   int getDisksNumber()
   {return where.length;}
}// class Configuration
